from enum import Enum, unique
from rdkit.Chem import Atom as RDAtom
from rdkit.Chem import Mol as RDMol
from rdkit.Chem import Bond as RDBond
from rdkit.Chem import BondType as RDBondType
from rdkit.Chem import (MolFromSmiles, MolToSmiles, MolToPDBBlock,
                        MolToMolBlock, GetPeriodicTable)
from openbabel import openbabel as ob
from openbabel.pybel import Molecule as PybelMol
from openbabel.pybel import readstring

from luna.wrappers.rdkit import new_mol_from_block
from luna.util.exceptions import (AtomObjectTypeError, BondObjectTypeError,
                                  IllegalArgumentError, MoleculeObjectError,
                                  MoleculeObjectTypeError)

import logging

logger = logging.getLogger()


@unique
class BondType(Enum):
    """An enumeration of bond types available at RDKit."""

    # Same values and bond types available at RDKit.
    UNSPECIFIED = 0
    SINGLE = 1
    DOUBLE = 2
    TRIPLE = 3
    QUADRUPLE = 4
    QUINTUPLE = 5
    HEXTUPLE = 6
    ONEANDAHALF = 7
    TWOANDAHALF = 8
    THREEANDAHALF = 9
    FOURANDAHALF = 10
    FIVEANDAHALF = 11
    AROMATIC = 12
    IONIC = 13
    HYDROGEN = 14
    THREECENTER = 15
    DATIVEONE = 16
    DATIVE = 17
    DATIVEL = 18
    DATIVER = 19
    OTHER = 20
    ZERO = 21


class OBBondType(Enum):
    """An enumeration of bond types available at Open Babel."""

    # Same values and bond types available at OpenBabel.
    SINGLE = 1
    DOUBLE = 2
    TRIPLE = 3
    AROMATIC = 5


class AtomWrapper:
    """This class provides util functions to access atomic properties and
    other information from RDKit and Open Babel objects.

    Parameters
    ----------
    atm_obj : :class:`AtomWrapper`, :class:`rdkit.Chem.rdchem.Atom`, or :class:`openbabel.OBAtom`
        An atom to wrap.
    mol_obj : `MolWrapper`, :class:`rdkit.Chem.rdchem.Mol`, :class:`openbabel.pybel.Molecule`, or None
        The molecule that contains the atom ``atom``. If None, the molecule is recovered directly from the atom object.

    Raises
    ------
    AtomObjectTypeError
        If the atom object is not an instance
            of `MolWrapper`, :class:`rdkit.Chem.rdchem.Mol`, or :class:`openbabel.pybel.Molecule`.
    """

    def __init__(self, atm_obj, mol_obj=None):
        if isinstance(atm_obj, self.__class__):
            atm_obj = atm_obj.unwrap()

        if not isinstance(atm_obj, RDAtom) and not isinstance(atm_obj, ob.OBAtom):
            logger.exception("Objects of type '%s' are not currently accepted." % atm_obj.__class__)
            raise AtomObjectTypeError("Objects of type '%s' are not currently accepted." % atm_obj.__class__)

        self._atm_obj = atm_obj
        self._parent = None

        if mol_obj is None:
            mol_obj = self.get_parent()
        self._parent = MolWrapper(mol_obj)

    @property
    def atm_obj(self):
        """:class:`rdkit.Chem.rdchem.Atom` or :class:`openbabel.OBAtom`: The wrapped atom object."""
        return self._atm_obj

    @atm_obj.setter
    def atm_obj(self, atm_obj):
        if not isinstance(atm_obj, RDAtom) and not isinstance(atm_obj, ob.OBAtom):
            logger.exception("Objects of type '%s' are not currently accepted." % atm_obj.__class__)
            raise AtomObjectTypeError("Objects of type '%s' are not currently accepted." % atm_obj.__class__)
        else:
            self._atm_obj = atm_obj

    @property
    def parent(self):
        """`MolWrapper`, :class:`rdkit.Chem.rdchem.Mol`, or :class:`openbabel.pybel.Molecule`: The molecule that contains this atom."""
        return self.get_parent()

    @parent.setter
    def parent(self, mol_obj):
        self._parent = MolWrapper(mol_obj)

    def get_idx(self):
        """Get this atom's internal index within a molecule.

        Returns
        -------
         : int
        """
        return self._atm_obj.GetIdx()  # Both RDKit and Open Babel have the same function name.

    def get_id(self):
        """Get this atom's unique id.

        When using RDKit, :py:meth:`get_id` and :py:meth:`get_idx` returns the same value.

        Returns
        -------
         : int
        """
        if self.is_rdkit_obj():
            return self._atm_obj.GetIdx()
        elif self.is_openbabel_obj():
            return self._atm_obj.GetId()

    def get_parent(self, wrapped=True):
        """Get the molecule that contains this atom.

        Parameters
        ----------
        wrapped : bool
            If True, wrap the molecule with :class:`~luna.wrappers.base.MolWrapper`.

        Returns
        -------
         : `MolWrapper`, :class:`rdkit.Chem.rdchem.Mol`, or :class:`openbabel.pybel.Molecule`
        """
        if self._parent is None:
            if self.is_rdkit_obj():
                mol_obj = self._atm_obj.GetOwningMol()
            elif self.is_openbabel_obj():
                mol_obj = self._atm_obj.GetParent()

            self._parent = mol_obj

        if wrapped:
            return MolWrapper(self._parent)
        return self._parent

    def get_atomic_num(self):
        """Get this atom's atomic number.

        Returns
        -------
         : int
        """
        
        # Both RDKit and Open Babel have the same function name.
        return self._atm_obj.GetAtomicNum()

    def get_symbol(self):
        """Get the element symbol of this atom.

        Returns
        -------
         : str
        """
        if self.is_rdkit_obj():
            return self._atm_obj.GetSymbol()
        elif self.is_openbabel_obj():
            return ob.GetSymbol(self._atm_obj.GetAtomicNum())

    def get_charge(self):
        """Get this atom's formal charge.

        Returns
        -------
         : int
        """
        return self._atm_obj.GetFormalCharge()  # Both RDKit and Open Babel have the same function name.

    def get_isotope(self):
        """Get this atom's isotope number.

        Returns
        -------
         : int
        """
        if self.is_rdkit_obj():
            return int(round(self._atm_obj.GetMass()))
        elif self.is_openbabel_obj():
            return int(round(self._atm_obj.GetExactMass()))

    def get_mass(self):
        """Get this atom's exact atomic mass, which can vary given the isotope.

        Returns
        -------
         : float
        """
        if self.is_rdkit_obj():
            return self._atm_obj.GetMass()
        elif self.is_openbabel_obj():
            return self._atm_obj.GetExactMass()

    def get_atomic_mass(self):
        """Get this atom's atomic mass given by standard IUPAC average molar mass.

        Returns
        -------
         : float
        """
        if self.is_rdkit_obj():
            return GetPeriodicTable().GetAtomicWeight(self.get_atomic_num())
        elif self.is_openbabel_obj():
            return self._atm_obj.GetAtomicMass()

    def get_neighbors_number(self, only_heavy_atoms=False):
        """Get the number of atoms bound to this atom.

        Parameters
        ----------
        only_heavy_atoms : bool
            If True, count only heavy atoms.

        Returns
        -------
         : int
        """

        # Subtract the number of hydrogens from the degree if only heavy atoms must be considered.
        penalty = 0 if only_heavy_atoms is False else self.get_h_count()
        return self.get_degree() - penalty

    def get_neighbors(self, wrapped=True):
        """Get all atoms that are bound to this atom.

        Parameters
        ----------
        wrapped : bool
            If True, wrap each atom with :class:`~luna.wrappers.base.AtomWrapper`.

        Returns
        -------
         : iterable of `AtomWrapper`, :py:class:`rdkit.Chem.rdchem.Atom`, or :class:`openbabel.OBAtom`
        """
        atoms = []
        if self.is_rdkit_obj():
            atoms = self._atm_obj.GetNeighbors()
        elif self.is_openbabel_obj():
            atoms = ob.OBAtomAtomIter(self._atm_obj)

        if atoms and wrapped:
            return [AtomWrapper(atom) for atom in atoms]
        return atoms

    def get_degree(self):
        """Get this atom's total degree.

        Returns
        -------
         : int
        """
        if self.is_rdkit_obj():
            return self._atm_obj.GetTotalDegree()
        elif self.is_openbabel_obj():
            return self._atm_obj.GetTotalDegree()

    def get_valence(self):
        """Get this atom's total valence (implicit and explicit)."""
        if self.is_rdkit_obj():
            return self._atm_obj.GetExplicitValence() + self._atm_obj.GetImplicitValence()
        elif self.is_openbabel_obj():
            return self._atm_obj.GetTotalValence()

            bonds = [b.GetBondOrder() for b in ob.OBAtomBondIter(self._atm_obj)]
            return sum(bonds) + self._atm_obj.ImplicitHydrogenCount()

    def get_h_count(self):
        """Get the total number of hydrogens (implicit and explicit) bound to this atom.

        Returns
        -------
         : int
        """
        if self.is_rdkit_obj():
            return self._atm_obj.GetTotalNumHs(includeNeighbors=True)
        elif self.is_openbabel_obj():
            return self._atm_obj.GetImplicitHCount() + self._atm_obj.ExplicitHydrogenCount()

    def get_bonds(self, wrapped=True):
        """Get this atom’s bonds.

        Parameters
        ----------
        wrapped : bool
            If True, wrap each bond with :class:`~luna.wrappers.base.BondWrapper`.

        Returns
        -------
         : iterable of `BondWrapper`, :class:`rdkit.Chem.rdchem.Bond`, or :class:`openbabel.OBBond`
        """
        bonds = []
        if self.is_rdkit_obj():
            bonds = self._atm_obj.GetBonds()
        elif self.is_openbabel_obj():
            bonds = ob.OBAtomBondIter(self._atm_obj)

        if bonds and wrapped:
            return [BondWrapper(bond) for bond in bonds]
        return bonds

    def get_atomic_invariants(self):
        """Get the atomic invariants of this atom.

        Atomic invariants are derived from ECFP [1]_ and E3FP [2]_ and consists of seven fields:

            * Number of heavy atoms;
            * Valence - Number of hydrogens;
            * Atomic number;
            * Isotope number;
            * Formal charge;
            * Number of hydrogens;
            * If the atom belongs to a ring or not.

        Returns
        -------
         : list
        """
        return [self.get_neighbors_number(only_heavy_atoms=True),       # Number of heavy atoms
                (self.get_valence() - self.get_h_count()),              # Valence - Num. Hs
                self.get_atomic_num(),                                  # Atomic number
                self.get_isotope(),                                     # Isotope number
                self.get_charge(),                                      # Formal charge
                self.get_h_count(),                                     # Num. Hs
                int(self.is_in_ring())]                                 # If the atom belongs to a ring or not

    def is_in_ring(self):
        """Check if this atom is in a ring."""
        return self._atm_obj.IsInRing()  # Both RDKit and Open Babel have the same function name.

    def is_aromatic(self):
        """Check if this atom is aromatic."""

        if self.is_rdkit_obj():
            return self._atm_obj.GetIsAromatic()
        elif self.is_openbabel_obj():
            return self._atm_obj.IsAromatic()

    def has_bond_type(self, bond_type):
        """Check if this atom has a bond of type ``bond_type``.

        Parameters
        ----------
        bond_type : `BondType`

        Raises
        ------
        IllegalArgumentError
            If the informed bond type is not an instance of `BondType`.
        """
        if isinstance(bond_type, BondType):
            return any([bond_obj.get_bond_type() == bond_type for bond_obj in self.get_bonds()])
        else:
            raise IllegalArgumentError("The informed bond type must be an instance of '%s'." % BondType)

    def has_only_bond_type(self, bond_type):
        """Check if this atom has only bonds of type ``bond_type``.

        Parameters
        ----------
        bond_type : `BondType`

        Raises
        ------
        IllegalArgumentError
            If the informed bond type is not an instance of `BondType`.
        """
        if isinstance(bond_type, BondType):
            return all([bond_obj.get_bond_type() == bond_type for bond_obj in self.get_bonds()])
        else:
            raise IllegalArgumentError("The informed bond type must be an instance of '%s'." % BondType)

    def set_charge(self, charge):
        """Set the formal charge of this atom.

        Parameters
        ----------
        charge : int
        """
        self._atm_obj.SetFormalCharge(charge)  # Both RDKit and Open Babel have the same function name.

    def set_as_aromatic(self, is_aromatic):
        """Set whether this atom is aromatic or not.

        Parameters
        ----------
        is_aromatic : bool
        """
        if self.is_rdkit_obj():
            self._atm_obj.SetIsAromatic(is_aromatic)
        elif self.is_openbabel_obj():
            self._atm_obj.SetAromatic(is_aromatic)

    def set_in_ring(self, in_ring):
        """Set whether this atom belongs to a ring or not.

        Parameters
        ----------
        in_ring : bool
        """
        if self.is_rdkit_obj():
            # TODO
            raise NotImplementedError("Currently, there is no function in RDKit to define if an atom belongs to a ring or not. "
                                      "Please, use Open Babel instead.")

        elif self.is_openbabel_obj():
            self._atm_obj.SetInRing(in_ring)

    def matches_smarts(self, smarts):
        """Check if this atom matches the substructure through a SMARTS substructure search.

        **Note:** currently, this function only works with molecules read with Open Babel.

        Parameters
        ----------
        smarts : str
            A substructure defined as SMARTS.

        Returns
        -------
         : bool or None
            Whether matches occurred. Return None if the molecule was read with RDKit.

        Examples
        --------
        First, let's read a molecule (glutamine) using Open Babel.

        >>> from luna.wrappers.base import MolWrapper
        >>> mol_obj = MolWrapper.from_smiles("N[C@@H](CCC(N)=O)C(O)=O", mol_obj_type="openbabel")

        Now, we'll loop over the list of atoms in the glutamine and check which atom is the amide's carbon.
        To do so, we can call the function :py:meth:`MolWrapper.get_atoms`, which, by default,
        returns `AtomWrapper` objects and then call :py:meth:`matches_smarts`.

        >>> for atm in mol_obj.get_atoms():
        >>>     print("%d\t%s\t%s" % (atm.get_idx(), atm.get_symbol(), atm.matches_smarts("C(N)(C)=O")))
        1   N   False
        2   C   False
        3   C   False
        4   C   False
        5   C   True
        6   N   False
        7   O   False
        8   C   False
        9   O   False
        10  O   False
        """
        if self.is_rdkit_obj():
            # TODO: Implement
            raise NotImplementedError("Currently, matches_smarts() does not support RDKit objects. "
                                      "Please, use Open Babel objects instead.")

        elif self.is_openbabel_obj():
            ob_smart = ob.OBSmartsPattern()
            ob_smart.Init(smarts)

            if ob_smart.Match(self.parent):
                for match in ob_smart.GetMapList():
                    if match[0] == self.get_idx():
                        return True
            return False

    def unwrap(self):
        """Return the original atomic object.

        Returns
        -------
         : :class:`rdkit.Chem.rdchem.Atom` or :class:`openbabel.OBAtom`
        """
        return self._atm_obj

    def is_rdkit_obj(self):
        """Check if this atom is an RDKit object."""
        if isinstance(self._atm_obj, RDAtom):
            return True
        return False

    def is_openbabel_obj(self):
        """Check if this atom is an Open Babel object."""
        if isinstance(self._atm_obj, ob.OBAtom):
            return True
        return False

    def __getattr__(self, attr):
        return getattr(self._atm_obj, attr)


class BondWrapper:
    """This class provides util functions to access bond properties and
    other information from RDKit and Open Babel objects.

    Parameters
    ----------
    bond_obj : `BondWrapper`, :py:class:`rdkit.Chem.rdchem.Bond`, or :py:class:`openbabel.OBBond`
        A bond to wrap.

    Raises
    ------
    BondObjectTypeError
        If the bond object is not an instance
            of `BondWrapper`, :class:`rdkit.Chem.rdchem.Bond`, or :class:`openbabel.pybel.OBBond`.
    """

    def __init__(self, bond_obj):
        if isinstance(bond_obj, self.__class__):
            bond_obj = bond_obj.unwrap()

        if not isinstance(bond_obj, RDBond) and not isinstance(bond_obj, ob.OBBond):
            logger.exception("Objects of type '%s' are not currently accepted." % bond_obj.__class__)
            raise BondObjectTypeError("Objects of type '%s' are not currently accepted." % bond_obj.__class__)

        self._bond_obj = bond_obj

    @property
    def bond_obj(self):
        """:class:`rdkit.Chem.rdchem.Bond` or :class:`openbabel.OBBond`: The wrapped bond object."""
        return self._bond_obj

    @bond_obj.setter
    def bond_obj(self, bond_obj):
        if not isinstance(bond_obj, RDBond) and not isinstance(bond_obj, ob.OBBond):
            logger.exception("Objects of type '%s' are not currently accepted." % bond_obj.__class__)
            raise AtomObjectTypeError("Objects of type '%s' are not currently accepted." % bond_obj.__class__)
        else:
            self._bond_obj = bond_obj

    def get_partner_atom(self, atm, wrapped=True):
        """Get the partner atom that forms this bond with ``atm``.

        Parameters
        ----------
        atm : `AtomWrapper`, :py:class:`rdkit.Chem.rdchem.Atom`, or :class:`openbabel.OBAtom`
            Get the partner of this atom.
        wrapped : bool
            If True, wrap the partner atom with `AtomWrapper`.

        Returns
        -------
         : `AtomWrapper`, :py:class:`rdkit.Chem.rdchem.Atom`, or :class:`openbabel.OBAtom`
        """
        if isinstance(atm, AtomWrapper):
            atm = atm.unwrap()

        partner = None
        if atm == self._bond_obj.GetBeginAtom():
            partner = self._bond_obj.GetEndAtom()
        elif atm == self._bond_obj.GetEndAtom():
            partner = self._bond_obj.GetBeginAtom()

        if partner and wrapped:
            return AtomWrapper(partner)
        return partner

    def get_begin_atom(self, wrapped=True):
        """Return the bond’s first atom.

        Parameters
        ----------
        wrapped : bool
            If True, wrap the atom with `AtomWrapper`.

        Returns
        -------
         : `AtomWrapper`, :py:class:`rdkit.Chem.rdchem.Atom`, or :class:`openbabel.OBAtom`
        """

        # Both RDKit and Open Babel have the same function name.
        if wrapped:
            return AtomWrapper(self._bond_obj.GetBeginAtom())
        return self._bond_obj.GetBeginAtom()

    def get_end_atom(self, wrapped=True):
        """Return the bond’s second atom.

        Parameters
        ----------
        wrapped : bool
            If True, wrap the atom with `AtomWrapper`.

        Returns
        -------
         : `AtomWrapper`, :py:class:`rdkit.Chem.rdchem.Atom`, or :class:`openbabel.OBAtom`
        """

        # Both RDKit and Open Babel have the same function name.
        if wrapped:
            return AtomWrapper(self._bond_obj.GetEndAtom())
        return self._bond_obj.GetEndAtom()

    def get_bond_type(self):
        """Get the bond type (e.g., single bond).

        Returns
        -------
         : `BondType`
        """
        if self.is_rdkit_obj():
            return BondType[self._bond_obj.GetBondType().name]
        elif self.is_openbabel_obj():
            # Map Open Babel bonds to BondType.
            return BondType[OBBondType(self._bond_obj.GetBondOrder()).name]

    def is_aromatic(self):
        """Check if this bond is aromatic or not."""
        if self.is_rdkit_obj():
            return self._bond_obj.GetIsAromatic()
        elif self.is_openbabel_obj():
            return self._bond_obj.IsAromatic()

    def set_bond_type(self, bond_type):
        """Set the type of the bond as a ``bond_type``.

        Parameters
        ----------
        bond_type : `BondType`
        """

        # Convert BondType to a valid type accepted by Open Babel or RDKit.
        if isinstance(bond_type, BondType):
            try:
                if self.is_rdkit_obj():
                    bond_type = RDBondType.values[bond_type.value]
                elif self.is_openbabel_obj():
                    bond_type = OBBondType[bond_type.name].value
            except KeyError as e:
                logger.exception(e)

                tool = "Open Babel" if self.is_openbabel_obj() else "RDKit"
                raise KeyError("The bond type '%s' is not a valid %s bond type." % (bond_type.name, tool))
        else:
            raise IllegalArgumentError("The informed bond type must be an instance of '%s'." % BondType)

        if self.is_rdkit_obj():
            try:
                self._bond_obj.SetBondType(bond_type)
            except Exception as e:
                logger.exception(e)
                raise
        elif self.is_openbabel_obj():
            self._bond_obj.SetBondOrder(bond_type)

    def set_as_aromatic(self, is_aromatic):
        """Set if this bond is aromatic or not.

        Parameters
        ----------
        is_aromatic : bool
        """
        if self.is_rdkit_obj():
            self._bond_obj.SetIsAromatic(is_aromatic)
        elif self.is_openbabel_obj():
            self._bond_obj.SetAromatic(is_aromatic)

    def unwrap(self):
        """Return the original bond object.

        Returns
        -------
         : :class:`rdkit.Chem.rdchem.Bond` or :class:`openbabel.OBBond`
        """
        return self._bond_obj

    def is_rdkit_obj(self):
        """Check if this bond is an RDKit object."""
        if isinstance(self._bond_obj, RDBond):
            return True
        return False

    def is_openbabel_obj(self):
        """Check if this bond is an Open Babel object."""
        if isinstance(self._bond_obj, ob.OBBond):
            return True
        return False

    def __getattr__(self, attr):
        return getattr(self._bond_obj, attr)


class MolWrapper:
    """This class provides util functions to access molecule properties and
    other information from RDKit and Open Babel objects.

    Parameters
    ----------
    mol_obj : `MolWrapper`, :class:`rdkit.Chem.rdchem.Mol`, or :class:`openbabel.pybel.Molecule`
        A molecule to wrap.

    Raises
    ------
    MoleculeObjectTypeError
        If the molecular object is not an instance
            of `MolWrapper`, :class:`rdkit.Chem.rdchem.Mol`, or :class:`openbabel.pybel.Molecule`.
    """

    def __init__(self, mol_obj):
        if isinstance(mol_obj, self.__class__):
            mol_obj = mol_obj.unwrap()
        elif isinstance(mol_obj, PybelMol):
            mol_obj = mol_obj.OBMol

        if not isinstance(mol_obj, RDMol) and not isinstance(mol_obj, ob.OBMol):
            logger.exception("Objects of type '%s' are not currently accepted." % mol_obj.__class__)
            raise MoleculeObjectTypeError("Objects of type '%s' are not currently accepted." % mol_obj.__class__)

        self._mol_obj = mol_obj

    @classmethod
    def from_mol_block(cls, block, mol_format, mol_obj_type="rdkit"):
        """Initialize a molecule from a string block.

        Parameters
        ----------
        block : str
            The molecular string block.
        mol_format : str
            Define the format in which the molecule is represented (e.g., 'mol2' or 'mol').
        mol_obj_type : {'rdkit', 'openbabel'}
            Define which library (RDKit or Open Babel) to use to parse the molecular block.
            The default value is RDKit.

        Returns
        -------
         : `MolWrapper`
        """
        if mol_obj_type == "rdkit":
            return cls(new_mol_from_block(block, mol_format))
        elif mol_obj_type == "openbabel":
            return cls(readstring(mol_format, block))

    @classmethod
    def from_smiles(cls, smiles, mol_obj_type="rdkit", name=None):
        """Initialize a molecule from a SMILES string.

        Parameters
        ----------
        smiles : str
            The SMILES string.
            Define the format in which the molecule is represented (e.g., 'mol2' or 'mol').
        mol_obj_type : {'rdkit', 'openbabel'}
            Define which library (RDKit or Open Babel) to use to parse the molecular block.
            The default value is RDKit.
        name : str, optional
            A name to identify the molecule.

        Returns
        -------
         : `MolWrapper`

        Raises
        ------
        MoleculeObjectError
            If it could not create a molecule from the provided SMILES.

        Examples
        --------

        >>> from luna.wrappers.base import MolWrapper
        >>> mol_obj = MolWrapper.from_smiles("N[C@@H](CCC(N)=O)C(O)=O", mol_obj_type="openbabel", name="Glutamine")

        """
        if mol_obj_type == "rdkit":
            mol = MolFromSmiles(smiles)
        elif mol_obj_type == "openbabel":
            mol = readstring("smi", smiles)

        # Raise an error if the molecule could not be created.
        if mol is None:
            raise MoleculeObjectError("It could not create a molecule from the provided SMILES '%s'." % smiles)

        mol = cls(mol)
        mol.set_name(name or smiles)

        return mol

    @property
    def mol_obj(self):
        """:class:`rdkit.Chem.rdchem.Mol` or :class:`openbabel.pybel.Molecule`: The wrapped molecular object."""
        return self._mol_obj

    @mol_obj.setter
    def mol_obj(self, mol_obj):
        if isinstance(mol_obj, PybelMol):
            mol_obj = mol_obj.OBMol

        if not isinstance(mol_obj, RDMol) and not isinstance(mol_obj, ob.OBMol):
            logger.exception("Objects of type '%s' are not currently accepted." % mol_obj.__class__)
            raise MoleculeObjectTypeError("Objects of type '%s' are not currently accepted." % mol_obj.__class__)

        self._mol_obj = mol_obj

    def as_rdkit(self):
        """If the molecule is an Open Babel object, convert it to an RDKit object."""
        if self.is_rdkit_obj():
            return self._mol_obj
        elif self.is_openbabel_obj():
            new_mol = MolWrapper.from_smiles(self.to_smiles(), mol_obj_type="rdkit")
            new_mol.set_name(self.get_name())
            return new_mol.unwrap()

    def as_openbabel(self):
        """If the molecule is an RDKit object, convert it to an Open Babel object."""
        if self.is_openbabel_obj():
            return self._mol_obj
        elif self.is_rdkit_obj():
            new_mol = MolWrapper.from_smiles(self.to_smiles(), mol_obj_type="openbabel")
            new_mol.set_name(self.get_name())
            return new_mol.unwrap()

    def get_name(self):
        """Get the molecule name.

        Returns
        -------
         : str
        """
        if self.is_rdkit_obj():
            if self._mol_obj.HasProp("_Name"):
                return self._mol_obj.GetProp("_Name")
            return ""
        elif self.is_openbabel_obj():
            return self._mol_obj.GetTitle()

    def get_atoms(self, wrapped=True):
        """Get all molecule's atoms.

        Parameters
        ----------
        wrapped : bool
            If True, wrap all atoms with `AtomWrapper`.

        Returns
        -------
         : iterable of `AtomWrapper`, :py:class:`rdkit.Chem.rdchem.Atom`, or :class:`openbabel.OBAtom`
        """
        atoms = []
        if self.is_rdkit_obj():
            atoms = self._mol_obj.GetAtoms()
        elif self.is_openbabel_obj():
            atoms = ob.OBMolAtomIter(self._mol_obj)

        if atoms and wrapped:
            return [AtomWrapper(atm, self) for atm in atoms]
        return atoms

    def get_bonds(self, wrapped=True):
        """Get all molecule's bonds.

        Parameters
        ----------
        wrapped : bool
            If True, wrap all bonds with `BondWrapper`.

        Returns
        -------
         : iterable of `BondWrapper`, :class:`rdkit.Chem.rdchem.Bond`, or :class:`openbabel.OBBond`
        """
        bonds = []
        if self.is_rdkit_obj():
            bonds = self._mol_obj.GetBonds()
        elif self.is_openbabel_obj():
            bonds = ob.OBMolBondIter(self._mol_obj)

        if bonds and wrapped:
            return [BondWrapper(bond) for bond in bonds]
        return bonds

    def get_num_heavy_atoms(self):
        """Get the number of heavy atoms in this molecule."""
        if self.is_rdkit_obj():
            return self._mol_obj.GetNumHeavyAtoms()
        elif self.is_openbabel_obj():
            return self._mol_obj.NumHvyAtoms()

    def get_atom_coord_by_id(self, atm_id):
        """Get the coordinates of an atom given by its id.

        Returns
        -------
         : array_like of float (size 3)
            Atomic coordinates (x, y, z).
        """
        if self.is_rdkit_obj():
            return list(self._mol_obj.GetConformer().GetAtomPosition(atm_id))
        elif self.is_openbabel_obj():
            atm = self._mol_obj.GetAtomById(atm_id)
            return [atm.GetX(), atm.GetY(), atm.GetZ()]

    def get_obj_type(self):
        """Get the object type ("rdkit" or "openbabel")."""
        if self.is_rdkit_obj():
            return "rdkit"
        elif self.is_openbabel_obj():
            return "openbabel"

    def has_name(self):
        """Check if this molecule has a name."""
        if self.get_name():
            return True
        return False

    def set_name(self, name):
        """Set a name to this molecule.

        Parameters
        ----------
        name : str
        """
        if self.is_rdkit_obj():
            self._mol_obj.SetProp("_Name", name)
        elif self.is_openbabel_obj():
            self._mol_obj.SetTitle(name)

    def to_smiles(self):
        """Return the canonical SMILES string for this molecule."""
        if self.is_rdkit_obj():
            return MolToSmiles(self._mol_obj)
        elif self.is_openbabel_obj():
            return PybelMol(self._mol_obj).write("smi").split("\t")[0]

    def to_pdb_block(self):
        """Return the PDB string block for this molecule."""
        if self.is_rdkit_obj():
            return MolToPDBBlock(self._mol_obj)
        elif self.is_openbabel_obj():
            return PybelMol(self._mol_obj).write("pdb")

    def to_mol_block(self):
        """Return the MOL string block for this molecule."""
        if self.is_rdkit_obj():
            return MolToMolBlock(self._mol_obj)
        elif self.is_openbabel_obj():
            return PybelMol(self._mol_obj).write("mol")

    def unwrap(self):
        """Return the original molecular object.

        Returns
        -------
         : :class:`rdkit.Chem.rdchem.Mol` or :class:`openbabel.pybel.Molecule`.
        """
        return self._mol_obj

    def is_rdkit_obj(self):
        """Check if this molecule is an RDKit object."""
        if isinstance(self._mol_obj, RDMol):
            return True
        return False

    def is_openbabel_obj(self):
        """Check if this molecule is an Open Babel object."""
        if isinstance(self._mol_obj, ob.OBMol):
            return True
        elif isinstance(self._mol_obj, PybelMol):
            return True
        return False

    def is_pybel_obj(self):
        """Check if this molecule is a Pybel object."""
        if isinstance(self._mol_obj, PybelMol):
            return True
        return False

    def __getattr__(self, attr):
        if self.mol_obj is None:
            return None
        return getattr(self._mol_obj, attr)

    def __getstate__(self):
        # Creates a copy of the class' dictionary in case we need to modify the molecular object to pickle it.
        my_dict = self.__dict__.copy()
        if self.is_openbabel_obj():
            my_dict["_mol_block"] = self.to_mol_block()
            my_dict["_mol_obj_type"] = self.get_obj_type()
            my_dict["_mol_obj"] = None
        return my_dict

    def __setstate__(self, state):
        if "_mol_obj_type" in state and "_mol_block" in state:
            state["_mol_obj"] = self.from_mol_block(state["_mol_block"], "mol", "openbabel").unwrap()
            del state["_mol_obj_type"]
            del state["_mol_block"]

        self.__dict__.update(state)



from luna.MyBio.PDB.PDBIO import Select


DEFAULT_ALTLOC = ("A", "1")


class Selector(Select):
    """Select atoms for PDB output.

    Parameters
    ----------
    keep_hydrog : bool
        If True (default), keeps all hydrogens. Otherwise, filter them out.
        Hydrogens are atoms whose element is either "H" or "D" (deuterium).
    keep_altloc : bool
        If True (default), keeps all atoms. Otherwise, keeps only atoms whose alternate location is in the list ``altloc``.
    altloc : iterable
        List of valid alternate location identifiers. The default valid values are "A" and "1".
    """

    def __init__(self, keep_hydrog=True, keep_altloc=True, altloc=DEFAULT_ALTLOC):
        self.keep_hydrog = keep_hydrog
        self.keep_altloc = keep_altloc
        self.altloc = altloc

    def accept_atom(self, atom):
        """Decide if the atom is valid or not."""

        # Hydrogen and Deuterium
        if not self.keep_hydrog and atom.element in ["H", "D"]:
            return False

        if self.keep_altloc:
            return True
        else:
            return not atom.is_disordered() or atom.get_altloc() in self.altloc


class ChainSelector(Selector):
    """Select atoms from chains defined at ``entries``.

    Parameters
    ----------
    entries : iterable of :class:`~luna.MyBio.PDB.Entity.Chain` instances
        Sequence of chain objects.
    """

    def __init__(self, entries, **kwargs):
        self.entries = entries
        super().__init__(**kwargs)

    def accept_chain(self, chain):
        """Decide if the chain is valid or not."""
        return True if (chain in self.entries) else False

    def accept_residue(self, res):
        """Decide if the residue is valid or not."""
        return self.accept_chain(res.get_parent())

    def accept_atom(self, atom):
        return super().accept_atom(atom) and self.accept_residue(atom.get_parent())


class ResidueSelector(Selector):
    """Select atoms from residues and other molecules defined at ``entries``.

    Parameters
    ----------
    entries : iterable of :class:`~luna.MyBio.PDB.Residue.Residue`
        Sequence of residue objects.
    """

    def __init__(self, entries, **kwargs):
        self.entries = entries
        super().__init__(**kwargs)

    def accept_residue(self, res):
        """Decide if the residue is valid or not."""
        return res in self.entries

    def accept_atom(self, atom):
        return super().accept_atom(atom) and self.accept_residue(atom.get_parent())


class ResidueSelectorBySeqNum(Selector):
    """Select atoms from compounds defined at ``entries``.

    Parameters
    ----------
    entries : iterable of int
        Sequence of residue sequence numbers.
    """

    def __init__(self, entries, **kwargs):
        self.entries = entries
        super().__init__(**kwargs)

    def accept_residue(self, res):
        """Decide if the residue is valid or not."""
        return True if (res.get_id()[1] in self.entries) else False

    def accept_atom(self, atom):
        return super().accept_atom(atom) and self.accept_residue(atom.get_parent())


class AtomSelector(Selector):
    """Select atoms defined at ``entries``.

    Parameters
    ----------
    entries : iterable of :class:`~luna.MyBio.PDB.Atom.Atom`
        Sequence of atom objects.
    """

    def __init__(self, entries, **kwargs):
        self.entries = entries
        super().__init__(**kwargs)

    def accept_atom(self, atom):
        return super().accept_atom(atom) and atom in self.entries



from luna.mol.entry import MolFileEntry
from luna.wrappers.pymol import PymolWrapper, PymolSessionManager
from luna.MyBio.util import entity_to_string
from luna.util.exceptions import PymolSessionNotInitialized
from luna.util.file import is_file_valid


import logging

logger = logging.getLogger()


class ShellViewer(PymolSessionManager):
    """Class that inherits from :class:`~luna.wrappers.pymol.PymolSessionManager` and implements :meth:`set_view`
    to depict shells (IFP features) and interactions in Pymol and save the view as a Pymol session.

    This class can be used to visualize multiple complexes into the same Pymol session, for instance, to compare
    binding modes and analyze similar shells. To do so, it is recommended that the protein structures are in the same coordinate system, i.e.,
    they should be aligned first. This can be achieved with :func:`luna.align.tmalign.align_2struct`
    or any other tool of your preference.

    Examples
    --------

    In the below example, we will assume a LUNA project object named ``proj_obj`` already exists.

    To visualize shells, we first need to generate them. So, let's define a
    :class:`~luna.interaction.fp.shell.ShellGenerator` object
    that will create shells over 2 iterations (levels). At each iteration, the shell radius will
    be increased by 3 and substructural information will be encoded following EIFP definition.
    Here, as an example, we will generate shells for the first :class:`~luna.mol.groups.AtomGroupsManager`
    object at ``proj_obj``.

    >>> from luna.interaction.fp.shell import ShellGenerator
    >>> from luna.interaction.fp.type import IFPType
    >>> num_levels, radius_step = 2, 3
    >>> sg = ShellGenerator(num_levels, radius_step, ifp_type=IFPType.EIFP)
    >>> atm_grps_mngr = list(proj_obj.atm_grps_mngrs)[0]
    >>> sm = sg.create_shells(atm_grps_mngr)

    The function :meth:`~luna.interaction.fp.shell.ShellGenerator.create_shells` returns a
    :class:`~luna.interaction.fp.shell.ShellManager` object, which provides built-in methods
    to access the created shells. Therefore, you can interact with it to select the shells you want to
    visualize in Pymol. As an example, let's select all unique shells at the last level:

    >>> shells = sm.get_shells_by_level(num_levels - 1, unique_shells=True)

    .. note::
        Levels are 0-indexed. So, the first level is 0, second is 1, etc.
        That means if ``num_levels`` is 5, the last level will be 4.

    As `ShellViewer` expects a list of tuples, we will define it now. The first item of the tuple
    is the :class:`~luna.mol.entry.Entry` instance, which represents a ligand. This can be obtained directly from
    the :class:`~luna.mol.groups.AtomGroupsManager` object. The second item is the list of shells you want to visualize.
    Finally, the third item can be either a PDB file or a directory.
    In this example, we will use the PDB directory defined during the LUNA project initialization.

    >>> shell_tuples = [(atm_grps_mngr.entry, shells, proj_obj.pdb_path)]

    To finish, we now create a new `ShellViewer` object and call
    :meth:`~luna.wrappers.pymol.PymolSessionManager.new_session`, which will initialize the session, depict shells and interactions, and save
    the session to an output PSE file.

    >>> from luna.interaction.fp.view import ShellViewer
    >>> sv = ShellViewer()
    >>> sv.new_session(shell_tuples, "example.pse")
    """

    def set_view(self, shell_tuples):
        """Depict shells (IFP features) and interactions into the current Pymol session.

        Parameters
        ----------
        shell_tuples : iterable of tuple
            Each tuple must contain three items: an :class:`~luna.mol.entry.Entry` instance,
            an iterable of :class:`~luna.interaction.fp.shell.Shell`,
            and a PDB file or the directory where the PDB file is located.
        """

        if not isinstance(self.wrapper, PymolWrapper):
            raise PymolSessionNotInitialized("No session was initialized.")

        for target_entry, shells, pathname in shell_tuples:

            if is_file_valid(pathname):
                pdb_file = pathname
            else:
                pdb_file = "%s/%s.pdb" % (pathname, target_entry.pdb_id)

            main_grp = target_entry.to_string(sep="-")

            mol_block = (entity_to_string(target_entry.get_biopython_structure())
                         if isinstance(target_entry, MolFileEntry) else None)

            # Load PDB and extract hetatm.
            self.load_pdb(pdb_file, main_grp, mol_block)

            residue_selections = set()
            for index, shell in enumerate(shells):
                sphere_obj = "%s.spheres.s%d_lvl%d_center%d" % (main_grp, index, shell.level, hash(shell.central_atm_grp))
                centroid_obj = "%s.centroid" % (sphere_obj)

                self.wrapper.add_pseudoatom(centroid_obj, {"color": "white", "pos": list(shell.central_atm_grp.centroid)})
                self.wrapper.hide([("nonbonded", centroid_obj)])
                self.wrapper.show([("spheres", centroid_obj), ("nb_spheres", centroid_obj), ("dots", centroid_obj)])
                self.wrapper.set("dot_color", "red")
                self.wrapper.set("sphere_scale", shell.radius, {"selection": centroid_obj})
                self.wrapper.set("sphere_transparency", 0.85, {"selection": centroid_obj})
                self.wrapper.run_cmds([("center", {"selection": centroid_obj})])

                self.wrapper.label([(centroid_obj, '"%s"' % "+".join([a.name for a in sorted(shell.central_atm_grp.atoms)]))])

                # Add interactions and styles.
                interacting_residue_sels = self.set_interactions_view(shell.interactions, main_grp, sphere_obj)

                residue_selections.update(interacting_residue_sels)

            if residue_selections:
                self.wrapper.select(name="%s.inter_residues" % main_grp, selection=" or ".join(residue_selections))

        self.set_last_details_to_view()



import argparse
import os
import shutil
import configparser

import sys
scripts = '/media/data/Workspace/Keiser lab/LUNA'
sys.path.append(scripts)


from luna.projects import LocalProject, EntryResults
from luna.version import __version__ as version
from luna.config.params import ProjectParams
from luna.interaction.view import InteractionViewer
from luna.interaction.fp.type import IFPType
from luna.util.file import create_directory
from luna.interaction.calc import InteractionsManager
from luna.util.logging import VERBOSITY_LEVEL

IFP_TYPES = {
    "EIFP": IFPType.EIFP,
    "FIFP": IFPType.FIFP,
    "HIFP": IFPType.HIFP,
}


class NegateAction(argparse.Action):
    def __call__(self, parser, ns, values, option):
        setattr(ns, self.dest, not option.startswith('--no-'))


class CountIFPAction(argparse.Action):
    def __call__(self, parser, ns, values, option):
        setattr(ns, self.dest, option == "-C")


def get_parser():
    parser = argparse.ArgumentParser()

    group = parser.add_argument_group('Entries')
    group.add_argument('-e', "--entries", type=str,
                       help="an input file containing a list of "
                            "entries to process.")
    group.add_argument("-p", "--prot", dest="pdb_id",
                       type=str,
                       help="the protein PDB id. "
                            "Mandatory if there is at least one entry "
                            "containing only the ligand name. "
                            "This only applies to ligands to be read "
                            "from a multimolecular file.")
    group.add_argument('-l', "--lig", dest="mol_file", type=str,
                       help="a molecular file containing 1 or more ligands. "
                            "Mandatory if there is at least one entry "
                            "containing only the ligand name. "
                            "This only applies to ligands to be read "
                            "from a multimolecular file.")
    group.add_argument('--esep', dest="entries_sep", type=str,
                       help="the entries separator. Default: ':'. "
                            "E.g.: 3QQK:A:X02:497.")
    group.add_argument('--fsep', dest="fields_sep", type=str,
                       help="the separator used to separate fields "
                            "(protein id, ligand name, molecular file, "
                            "multimol flag) in <ENTRIES_FILE>. "
                            "Default: ','.")

    group = parser.add_argument_group('Paths')
    group.add_argument('-w', '--workdir', dest="working_path", type=str,
                       help="the path where the project and "
                            "its results will be saved.")
    group.add_argument('--pdbdir', dest="pdb_path", type=str,
                       help="path containing local PDB files or to where the "
                            "PDB files will be downloaded. Default: None.")
    group.add_argument('--overwrite', '--no-overwrite', dest="overwrite_path",
                       action=NegateAction, nargs=0,
                       help="overwrite or not an existing project. "
                            " Default: None.")

    group = parser.add_argument_group('Hydrogens')
    group.add_argument('--addh', '--no-addh', dest="add_h",
                       action=NegateAction, nargs=0,
                       help="turn off hydrogen addition.")
    group.add_argument('--ph', type=float,
                       help="add hydrogens appropriate for pH. Default: 7.4.")

    group = parser.add_argument_group('Standardization')
    group.add_argument('--amend', '--no-amend', dest="amend_mol",
                       action=NegateAction, nargs=0,
                       help="fix atomic charges, valence, and bond "
                            "types for small molecules and residues at "
                            "PDB files.")

    group = parser.add_argument_group('Features')
    group.add_argument('-F', '--feat', dest="feat_cfg", type=str,
                       help="a feature definition file (FDef) containing "
                            "chemical and pharmacophoric rules. "
                            "If not provided, the default LUNA features "
                            "definition will be used instead.")

    group = parser.add_argument_group('Interactions')
    group.add_argument('-I', '--inter', dest="inter_cfg", type=str,
                       help="a config file defining interaction parameters. "
                            "If not provided, the default LUNA config "
                            "file will be used instead.")
    group.add_argument('--noncov', '--no-noncov', dest="add_non_cov",
                       action=NegateAction, nargs=0,
                       help="turn on/off non-covalent interactions.")
    group.add_argument('--cov', '--no-cov', dest="add_cov",
                       action=NegateAction, nargs=0,
                       help="turn on/off covalent interactions.")
    group.add_argument('--prox', '--no-prox', dest="add_proximal",
                       action=NegateAction, nargs=0,
                       help="turn on/off proximal contacts.")
    group.add_argument('--aa', '--no-aa', dest="add_atom_atom",
                       action=NegateAction, nargs=0,
                       help="turn on/off atom-atom interactions.")
    group.add_argument('--adi', '--no-adi', dest="add_dependent_inter",
                       action=NegateAction, nargs=0,
                       help="turn on/off dependent interactions.")
    group.add_argument('--awp', '--no-awp',
                       dest="add_h2o_pairs_with_no_target",
                       action=NegateAction, nargs=0,
                       help="accept or not water that doesn't interact "
                            "with the target ligand.")
    group.add_argument('--sdr', '--no-sdr', dest="strict_donor_rules",
                       action=NegateAction, nargs=0,
                       help="turn on/off strict donor rules.")
    group.add_argument('--swdr', '--no-swdr', dest="strict_weak_donor_rules",
                       action=NegateAction, nargs=0,
                       help="turn on/off strict weak donor rules.")
    group.add_argument('--lazy', dest="lazy_comps_list", type=str,
                       help="a comma-separated list of molecule names "
                            "(PDB ids) to ignore explicit hydrogen "
                            "position during the calculation of (weak) "
                            "hydrogen bonds. "
                            "Default: HOH,DOD,WAT,H2O,OH2,NH3,NH4.")

    group = parser.add_argument_group('Interactions filter')
    action = group.add_mutually_exclusive_group()
    action.add_argument('--filter', dest="filter_cfg", type=str,
                        help="a config file defining interaction filters. "
                             "If not provided, all interaction pairs will be "
                             "accepted by default.")

    action.add_argument('--df', dest="default_filter", type=str,
                        choices=['pli', 'ppi', 'pni', 'nni', 'nli'],
                        help="use default filters for interactions, where "
                             "pli, ppi, pni, nni, and nli stand for "
                             "protein-ligand, protein-protein, "
                             "protein-nucleotide, nucleotide-nucleotide, and "
                             "nucleotide-ligand interactions respectively")

    group = parser.add_argument_group('Binding mode filters')
    group.add_argument('--bind', dest="bind_cfg", type=str,
                       help="a config file defining binding mode rules "
                            "to filter interactions.")

    group = parser.add_argument_group('LUNA interaction fingerprints')
    group.add_argument('--ifp', '--no-ifp', dest="ifp",
                       action=NegateAction, nargs=0,
                       help="generate or not LUNA interaction fingerprints.")
    group.add_argument('-L', dest="ifp_num_levels", type=int,
                       help="the number of level defines the number of "
                            "iterations to construct the fingerprint. "
                            "Default: 2.")
    group.add_argument('-R', dest="ifp_radius_step",
                       type=float,
                       help="the radius growth rate defines the multiplier "
                            "to increase the sphere size at each level. "
                            "Default: 5.73171.")
    group.add_argument('-S', dest="ifp_length", type=int,
                       help="the fingerprint length. Default: 4096.")
    group.add_argument('-B', '-C', dest="ifp_count",
                       action=CountIFPAction, nargs=0,
                       help="use bit fingerprints. The default value "
                            "is False, which imply that count fingerprints "
                            "are used instead.")
    group.add_argument('-T', dest="ifp_type", type=str,
                       choices=['EIFP', 'HIFP', 'FIFP'],
                       help="the fingerprint type. Default: EIFP.")
    group.add_argument('--diff', '--no-diff', dest="ifp_diff_comp_classes",
                       action=NegateAction, nargs=0,
                       help="turn on/off differentiation between "
                            "compound classes.")
    group.add_argument('--ifp-out', dest="ifp_output", type=str,
                       help="the fingerprint output file. "
                            "Default: <WORKING_PATH>/results/"
                            "fingerprints/ifp.csv.")
    group.add_argument('--ifp-matrix', dest="ifp_sim_matrix_output",
                       type=str, nargs='?', const=True,
                       help="if provided, save a similarity "
                            "matrix to <IFP_MATRIX>. "
                            "Optionally, use the flag without "
                            "parameters to save the matrix at the "
                            "default path: <WORKING_PATH>/results/"
                            "fingerprints/ifp_sim_matrx.csv")

    group = parser.add_argument_group('Molecular fingerprints')
    group.add_argument('--mfp', '--no-mfp', dest="mfp",
                       action=NegateAction, nargs=0,
                       help="generate or not molecular fingerprints.")
    group.add_argument('--mfp-out', dest="mfp_output", type=str,
                       help="the molecular fingerprint output file. "
                            "Default: <WORKING_PATH>/results/"
                            "fingerprints/mfp.csv.")

    group = parser.add_argument_group('Pymol sessions')
    group.add_argument('--pse', '--no-pse', dest="pse",
                       action=NegateAction, nargs=0,
                       help="export or not interactions to Pymol.")
    group.add_argument('--psedir', dest="pse_path", type=str,
                       help="save PSE files to <PSE_PATH>. "
                            "Default: <WORKING_PATH>/results/pse/.")

    # General options
    parser.add_argument('--lib', dest="mol_obj_type", type=str,
                        choices=['rdkit', 'openbabel'],
                        help="which library (RDKit or Open Babel) to use "
                             "to parse molecules. Default: rdkit.")
    parser.add_argument('--append', '--no-append', dest="append_mode",
                        action=NegateAction, nargs=0,
                        help="skip or not entries when a result for them "
                             "already exists in <WORKING_PATH>.")
    parser.add_argument('--cache', '--no-cache', dest="use_cache",
                        action=NegateAction, nargs=0,
                        help="cache or not protein information"
                             "to save processing time.")
    parser.add_argument("-v", dest="verbosity", type=int,
                        choices=sorted(VERBOSITY_LEVEL.keys()),
                        help="verbosity level. Default: 3.")
    parser.add_argument('--log', '--no-log', dest="logging_enabled",
                        action=NegateAction, nargs=0,
                        help="turn on/off the logging system.")
    parser.add_argument("--nproc", type=int,
                        help="the number of processors to use. "
                             "Default: -1 (use all available CPUs).")
    parser.add_argument('--fork', dest="fork_project", type=str,
                        help="if provided, copy an existing "
                             "project to <WORKING_PATH>.")
    parser.add_argument('-c', '--conf', dest="luna_config", type=str,
                        help="load LUNA parameters from the "
                             "provided config file.")

    return parser


def main():

    conf_parser = argparse.ArgumentParser(add_help=False)
    conf_parser.add_argument("-c", "--conf", dest="luna_config", type=str,
                             help="Load LUNA parameters from the "
                                  "provided config file.")
    args, remaining_argv = conf_parser.parse_known_args()

    # Read the provided config file and set its values to 'config_values'.
    # This dict will then be used to overwrite the default valuies.
    config_values = {}
    if args.luna_config:
        config = configparser.ConfigParser()
        config.read([args.luna_config])

        for section in config.sections():
            config_values.update(dict(config.items(section)))

    # Get the list of expected arguments.
    parser = get_parser()
    # Set default values with parameters provided in the config file.
    parser.set_defaults(**config_values)
    # Overwrite all parameters using values provided through command line.
    args = parser.parse_args(remaining_argv)

    params = dict(args._get_kwargs())

    # Remove all, but params provided through a config file or
    # through command line.
    params = {k: v for k, v in params.items()
              if k in config_values or v is not None}

    # If provided without parameter, this argument will be set to True.
    # That means the default directory should be used instead.
    # However, ProjectParams expects a string. To avoid raising an error,
    # we remove it from 'params' to deal with this argument separatelly.
    ifp_sim_matrix_output = params.pop("ifp_sim_matrix_output", None)

    # Parse params.
    # It will load in default values for every parameter not provided.
    params = ProjectParams(params, fill_defaults=True)

    required = []
    if params.get("entries", None) is None:
        required.append("-e/--entries")

    if params.get("working_path", None) is None:
        required.append("-w/--workdir")

    if required:
        parser.error("the following arguments are required: %s. "
                     "It can also be provided in a config file (-c/--conf)."
                     % ", ".join(required))

    # Fork a project
    if args.fork_project:
        fork_proj_pkl_file = LocalProject.get_project_file(args.fork_project)
        if not os.path.exists(fork_proj_pkl_file):
            raise OSError("No valid project was found at '%s'"
                          % args.fork_project)
        shutil.copytree(args.fork_project, params["working_path"])

    # As the same project can be reutilized using different IFP parameters
    # and binding mode filters, we remove the below args from 'params'
    # so that we first process or load a project, and only then we apply the
    # binding mode filter, calculate IFPs, and PSE sessions.
    binding_mode_filter = params.get("binding_mode_filter", None)
    calc_mfp = params.get("calc_mfp", False)
    calc_ifp = params.get("calc_ifp", False)
    out_pse = params.get("out_pse", False)
    out_pse = params.get("out_pse", False)

    # Set those parameters again using its turned off values.
    params["binding_mode_filter"] = None
    params["calc_mfp"] = False
    params["calc_ifp"] = False
    params["out_pse"] = False

    proj_pkl_file = LocalProject.get_project_file(params["working_path"])
    if not os.path.exists(proj_pkl_file) or params["overwrite_path"]:
        print(u"\u25a8 Creating a new LUNA project: '%s'...\n"
              % params["working_path"])

        pli_obj = LocalProject(**params)
        pli_obj.run()
    else:
        print(u"\u25a8 Reloading existing project: '%s'...\n"
              % params["working_path"])
        pli_obj = LocalProject.load(proj_pkl_file,
                                    logging_enabled=params["logging_enabled"])

    if binding_mode_filter:
        pli_obj.binding_mode_filter = binding_mode_filter

        for agm in pli_obj.atm_grps_mngrs:
            im = InteractionsManager(agm.get_all_interactions(), agm.entry)
            im.filter_out_by_binding_mode(binding_mode_filter)

            entry_results = EntryResults(agm.entry, agm, im, None, None)
            pkl_file = "%s/chunks/%s.pkl.gz" % (pli_obj.working_path,
                                                agm.entry.to_string())
            entry_results.save(pkl_file)

            csv_file = ("%s/results/interactions/%s.csv"
                        % (pli_obj.working_path, agm.entry.to_string()))
            im.to_csv(csv_file)

    if calc_ifp or calc_mfp:
        if calc_ifp:
            ifp_output = params["ifp_output"]
            if ifp_output is None:
                create_directory("%s/results/fingerprints"
                                 % params["working_path"])
                ifp_output = ("%s/results/fingerprints/ifp.csv"
                              % params["working_path"])

            pli_obj.calc_ifp = True
            pli_obj.ifp_num_levels = params["ifp_num_levels"]
            pli_obj.ifp_radius_step = params["ifp_radius_step"]
            pli_obj.ifp_length = params["ifp_length"]
            pli_obj.ifp_type = params["ifp_type"]
            pli_obj.ifp_diff_comp_classes = params["ifp_diff_comp_classes"]
            pli_obj.ifp_count = params["ifp_count"]
            pli_obj.ifp_output = ifp_output

            if ifp_sim_matrix_output:
                if ifp_sim_matrix_output is True:
                    default_output = ("%s/results/fingerprints/"
                                      "ifp_sim_matrix.csv"
                                      % pli_obj.working_path)
                    ifp_sim_matrix_output = default_output

                pli_obj.ifp_sim_matrix_output = ifp_sim_matrix_output

        if calc_mfp:
            mfp_output = params["mfp_output"]
            if mfp_output is None:
                create_directory("%s/results/fingerprints"
                                 % params["working_path"])
                mfp_output = ("%s/results/fingerprints/mfp.csv"
                              % params["working_path"])
            pli_obj.calc_mfp = True

        pli_obj.generate_fps()

    if out_pse:
        pli_obj.out_pse = True

        pli_obj.pse_path = (params["pse_path"]
                            or f"{pli_obj.working_path}/results/pse")
        create_directory(pli_obj.pse_path)

        for ic in pli_obj.interactions_mngrs:
            pse_file = f"{pli_obj.pse_path}/%s.pse" % ic.entry.to_string()
            viewer = InteractionViewer(add_directional_arrows=False)
            viewer.new_session([(ic.entry, ic, pli_obj.pdb_path)], pse_file)

    pli_obj.save_config_file()


if __name__ == '__main__':
    print()
    main()
